<?php

/**
 * @file
 * Generate barcodes.
 */

/**
 * Class Barcode.
 */
class DrawBarcode {
  /**
   * GD library function.
   *
   * @param object $res
   *   Result variable.
   * @param string $color
   *   Color variable.
   * @param int $x
   *   X coordinate variable.
   * @param int $y
   *   Y coordinate variable.
   * @param int $angle
   *   Angle variable.
   * @param string $type
   *   Type variable.
   * @param string $datas
   *   Barcode data variable.
   * @param int $width
   *   Barcode with variable.
   * @param int $height
   *   Barcode height variable.
   *
   * @return multitypestring
   *   Return the image property.
   */
  public static function gd($res, $color, $x, $y, $angle, $type, $datas, $width = NULL, $height = NULL) {
    return self::drawBarcodeImage(__FUNCTION__, $res, $color, $x, $y, $angle, $type, $datas, $width, $height);
  }

  /**
   * FPDF library function.
   *
   * @param object $res
   *   Result variable.
   * @param string $color
   *   Color variable.
   * @param int $x
   *   X coordinate variable.
   * @param int $y
   *   Y coordinate variable.
   * @param int $angle
   *   Angle variable.
   * @param string $type
   *   Type variable.
   * @param string $datas
   *   Barcode data variable.
   * @param int $width
   *   Barcode with variable.
   * @param int $height
   *   Barcode height variable.
   *
   * @return multitypestring
   *   Return the image property.
   */
  public static function fpdf($res, $color, $x, $y, $angle, $type, $datas, $width = NULL, $height = NULL) {
    return self::drawBarcodeImage(__FUNCTION__, $res, $color, $x, $y, $angle, $type, $datas, $width, $height);
  }

  /**
   * Draw Barcode pattern.
   *
   * @param object $call
   *   Call variable.
   * @param object $res
   *   Result variable.
   * @param string $color
   *   Color variable.
   * @param int $x
   *   X coordinate variable.
   * @param int $y
   *   Y coordinate variable.
   * @param int $angle
   *   Angle variable.
   * @param string $type
   *   Type variable.
   * @param string $datas
   *   Barcode data variable.
   * @param int $width
   *   Barcode with variable.
   * @param int $height
   *   Barcode height variable.
   *
   * @return multitypestring
   *   Return the image property.
   */
  private static function drawBarcodeImage($call, $res, $color, $x, $y, $angle, $type, $datas, $width, $height) {
    $digit = '';
    $hri = '';
    $code = '';
    $crc = TRUE;
    $rect = FALSE;
    $b2d = FALSE;
    if (is_array($datas)) {
      foreach (array('code' => '', 'crc' => TRUE, 'rect' => FALSE) as $v => $def) {
        $$v = isset($datas[$v]) ? $datas[$v] : $def;
      }
      $code = $code;
    }
    else {
      $code = $datas;
    }
    if ($code == '') {
      return FALSE;
    }
    $code = (string) $code;
    $type = strtolower($type);
    switch ($type) {
      case 'std25':
      case 'int25':
        $digit = DrawBarcodeModelI25::getDigit($code, $crc, $type);
        $hri = DrawBarcodeModelI25::compute($code, $crc, $type);
        break;

      case 'ean8':
      case 'ean13':
        $digit = DrawBarcodeModelEAN::getDigit($code, $type);
        $hri = DrawBarcodeModelEAN::compute($code, $type);
        break;

      case 'code11':
        $digit = DrawBarcodeModel11::getDigit($code);
        $hri = $code;
        break;

      case 'code39':
        $digit = DrawBarcodeModel39::getDigit($code);
        $hri = $code;
        break;

      case 'code93':
        $digit = DrawBarcodeModel93::getDigit($code, $crc);
        $hri = $code;
        break;

      case 'code128':
        $digit = DrawBarcodeModel128::getDigit($code);
        $hri = $code;
        break;

      case 'codabar':
        $digit = DrawBarcodeModelCodabar::getDigit($code);
        $hri = $code;
        break;

      case 'msi':
        $digit = DrawBarcodeModelMSI::getDigit($code, $crc);
        $hri = DrawBarcodeModelMSI::compute($code, $crc);
        break;

      case 'datamatrix':
        $digit = DrawBarcodeModelDatamatrix::getDigit($code, $rect);
        $hri = $code;
        $b2d = TRUE;
        break;
    }
    if ($digit == '') {
      return FALSE;
    }
    if ($b2d) {
      $width = is_null($width) ? 5 : $width;
      $height = $width;
    }
    else {
      $width = is_null($width) ? 1 : $width;
      $height = is_null($height) ? 50 : $height;
      $digit = self::bitStringTo2Darray($digit);
    }
    if ($call == 'gd') {
      $result = self::digitToGdRenderer($res, $color, $x, $y, $angle, $width, $height, $digit);
    }
    elseif ($call == 'fpdf') {
      $result = self::digitToFpdfRenderer($res, $color, $x, $y, $angle, $width, $height, $digit);
    }
    $result['hri'] = $hri;
    return $result;
  }

  /**
   * Convert a bit string to an array of array of bit char.
   *
   * @param int $digit
   *   Barcode dight variable.
   *
   * @return multitypestring
   *   Return array string value.
   */
  private static function bitStringTo2Darray($digit) {
    $d = array();
    $len = strlen($digit);
    for ($i = 0; $i < $len; $i++) {
      $d[$i] = $digit[$i];
    }
    return (array($d));
  }

  /**
   * GD barcode renderer.
   *
   * @param object $gd
   *   Gd library variable.
   * @param string $color
   *   Color variable.
   * @param int $xi
   *   Xi - x coordinate variable.
   * @param int $yi
   *   Yi - y coordinate variable.
   * @param int $angle
   *   Angle variable.
   * @param int $mw
   *   Width variable.
   * @param int $mh
   *   Height variable.
   * @param int $digit
   *   Barcode dight variable.
   *
   * @return multitypeint
   *   Return integer array.
   */
  private static function digitToGdRenderer($gd, $color, $xi, $yi, $angle, $mw, $mh, $digit) {
    $lines = count($digit);
    $columns = count($digit[0]);
    $angle = deg2rad(-$angle);
    $cos = cos($angle);
    $sin = sin($angle);
    $x = 0;
    $y = 0;
    $x1 = 0;
    $x2 = 0;
    $y1 = 0;
    $y2 = 0;
    self::rotate($columns * $mw / 2, $lines * $mh / 2, $cos, $sin, $x, $y);
    $xi -= $x;
    $yi -= $y;

    for ($y = 0; $y < $lines; $y++) {
      $len = 0;
      $current = $digit[$y][0];
      for ($x = 0; $x < $columns; $x++) {
        if ($current == $digit[$y][$x]) {
          $len++;
        }
        else {
          if ($current == '1') {
            $px = $len * $mw;
            $xt = $xi + ($x - $len) * $mw;
            if ($angle == 0) {
              if ($px > 2) {
                imagefilledrectangle($gd, $xt, $yi + $y * $mh, $xt + $px - 1, $yi + ($y + 1) * $mh, $color);
              }
              else {
                for ($i = 0; $i < $px; $i++) {
                  imageline($gd, $xt + $i, $yi + $y * $mh, $xt + $i, $yi + ($y + 1) * $mh, $color);
                }
              }
            }
            else {
              for ($i = 0; $i < $px; $i++) {
                self::rotate($xt + $i - $xi, $y * $mh, $cos, $sin, $x1, $y1);
                self::rotate($xt + $i - $xi, ($y + 1) * $mh, $cos, $sin, $x2, $y2);
                imageline($gd, $xi + $x1, $yi + $y1, $xi + $x2, $yi + $y2, $color);
              }
            }
          }
          $current = $digit[$y][$x];
          $len = 1;
        }
      }
      if (($len > 0) && ($current == '1')) {
        $px = $len * $mw;
        $xt = $xi + ($columns - $len) * $mw;
        $y = $lines - 1;
        if ($angle == 0) {
          if ($px > 2) {
            imagefilledrectangle($gd, $xt, $yi + $y * $mh, $xt + $px - 1, $yi + ($y + 1) * $mh, $color);
          }
          else {
            for ($i = 0; $i < $px; $i++) {
              imageline($gd, $xt + $i, $yi + $y * $mh, $xt + $i, $yi + ($y + 1) * $mh, $color);
            }
          }
        }
        else {
          for ($i = 0; $i < $px; $i++) {
            self::rotate($xt + $i - $xi, $y * $mh, $cos, $sin, $x1, $y1);
            self::rotate($xt + $i - $xi, ($y + 1) * $mh, $cos, $sin, $x2, $y2);
            imageline($gd, $xi + $x1, $yi + $y1, $xi + $x2, $yi + $y2, $color);
          }
        }
      }
    }
    return self::result($xi, $yi, $columns, $lines, $mw, $mh, $cos, $sin);
  }

  /**
   * GD barcode renderer.
   *
   * @param object $pdf
   *   Pdf variable.
   * @param string $color
   *   Color variable.
   * @param int $xi
   *   X coordinate variable.
   * @param int $yi
   *   Y coordinate variable.
   * @param int $angle
   *   Angle variable.
   * @param int $mw
   *   Width variable.
   * @param int $mh
   *   Height variable.
   * @param int $digit
   *   Barcode digit variable.
   *
   * @return multitypeint
   *   Return integer array.
   */
  private static function digitToFpdfRenderer($pdf, $color, $xi, $yi, $angle, $mw, $mh, $digit) {
    $lines = count($digit);
    $columns = count($digit[0]);
    $angle = deg2rad(-$angle);
    $cos = cos($angle);
    $sin = sin($angle);
    $x = 0;
    $y = 0;
    $x1 = 0;
    $x2 = 0;
    $y1 = 0;
    $y2 = 0;
    $pdf->SetLineWidth($mw);

    if (!is_array($color)) {
      if (preg_match('`([0-9A-F]{2})([0-9A-F]{2})([0-9A-F]{2})`i', $color, $m)) {
        $color = array(hexdec($m[1]), hexdec($m[2]), hexdec($m[3]));
      }
      else {
        $color = array(0, 0, 0);
      }
    }
    $color = array_values($color);
    $pdf->SetDrawColor($color[0], $color[1], $color[2]);
    $pdf->SetFillColor($color[0], $color[1], $color[2]);

    self::rotate($columns * $mw / 2, $lines * $mh / 2, $cos, $sin, $x, $y);
    $xi -= $x;
    $yi -= $y;

    for ($y = 0; $y < $lines; $y++) {
      $len = 0;
      $current = $digit[$y][0];
      for ($x = 0; $x < $columns; $x++) {
        if ($current == $digit[$y][$x]) {
          $len++;
        }
        else {
          if ($current == '1') {
            $px = $len;
            $xt = $xi + ($x - $len) * $mw;
            if ($angle == 0) {
              if ($px > 2) {
                $pdf->rect($xt, $yi + $y * $mh, $px - 1, $mh, 'F');
              }
              else {
                for ($i = 0; $i < $px; $i++) {
                  $pdf->line($xt + $i, $yi + $y * $mh, $xt + $i, $yi + ($y + 1) * $mh);
                }
              }
            }
            else {
              for ($i = 0; $i < $px; $i++) {
                self::rotate($xt + $i - $xi, $y * $mh, $cos, $sin, $x1, $y1);
                self::rotate($xt + $i - $xi, ($y + 1) * $mh, $cos, $sin, $x2, $y2);
                $pdf->line($xi + $x1, $yi + $y1, $xi + $x2, $yi + $y2);
              }
            }
          }
          $current = $digit[$y][$x];
          $len = 1;
        }
      }
      if (($len > 0) && ($current == '1')) {
        $px = $len;
        $xt = $xi + ($columns - $len) * $mw;
        $y = $lines - 1;
        if ($angle == 0) {
          if ($px > 2) {
            $pdf->rect($xt, $yi + $y * $mh, $px - 1, $mh, 'F');
          }
          else {
            for ($i = 0; $i < $px; $i++) {
              $pdf->line($xt + $i, $yi + $y * $mh, $xt + $i, $yi + ($y + 1) * $mh);
            }
          }
        }
        else {
          for ($i = 0; $i < $px; $i++) {
            self::rotate($xt + $i - $xi, $y * $mh, $cos, $sin, $x1, $y1);
            self::rotate($xt + $i - $xi, ($y + 1) * $mh, $cos, $sin, $x2, $y2);
            $pdf->line($xi + $x1, $yi + $y1, $xi + $x2, $yi + $y2);
          }
        }
      }
    }
    return self::result($xi, $yi, $columns, $lines, $mw, $mh, $cos, $sin);
  }

  /**
   * Return the barcode finalized image.
   *
   * @param int $xi
   *   X coordinate variable.
   * @param int $yi
   *   Y cooridnate variable.
   * @param int $columns
   *   Column variable.
   * @param int $lines
   *   Lines variable.
   * @param int $mw
   *   Width variable.
   * @param int $mh
   *   Height variable.
   * @param int $cos
   *   Cos variable.
   * @param int $sin
   *   Sin variable.
   *
   * @return multitypeint
   *   Return integer array.
   */
  private static function result($xi, $yi, $columns, $lines, $mw, $mh, $cos, $sin) {
    $x1 = $xi;
    $y1 = $yi;
    $x2 = $xi;
    $y2 = $yi;
    $x3 = $xi;
    $y3 = $yi;
    $x4 = $xi;
    $y4 = $yi;
    self::rotate(0, 0, $cos, $sin, $x1, $y1);
    self::rotate($columns * $mw, 0, $cos, $sin, $x2, $y2);
    self::rotate($columns * $mw, $lines * $mh, $cos, $sin, $x3, $y3);
    self::rotate(0, $lines * $mh, $cos, $sin, $x4, $y4);
    return array(
      'width' => $columns * $mw,
      'height' => $lines * $mh,
      'p1' => array(
        'x' => $xi + $x1,
        'y' => $yi + $y1,
      ),
      'p2' => array(
        'x' => $xi + $x2,
        'y' => $yi + $y2,
      ),
      'p3' => array(
        'x' => $xi + $x3,
        'y' => $yi + $y3,
      ),
      'p4' => array(
        'x' => $xi + $x4,
        'y' => $yi + $y4,
      ),
    );
  }

  /**
   * Rotate the barcode image by value.
   *
   * @param int $x1
   *   X cooridnate variable.
   * @param int $y1
   *   Y cooridnate variable.
   * @param int $cos
   *   Cos variable.
   * @param int $sin
   *   Sin variable.
   * @param int $x
   *   X coordinate variable.
   * @param int $y
   *   Y coordinate variable.
   */
  private static function rotate($x1, $y1, $cos, $sin, &$x, &$y) {
    $x = $x1 * $cos - $y1 * $sin;
    $y = $x1 * $sin + $y1 * $cos;
  }

  /**
   * Rotate the barcode image by angle.
   *
   * @param int $x1
   *   X cooridnate variable.
   * @param int $y1
   *   Y cooridnate variable.
   * @param int $angle
   *   Angle variable.
   * @param int $x
   *   X cooridnate variable.
   * @param int $y
   *   Y cooridnate variable.
   */
  public static function rotates($x1, $y1, $angle, &$x, &$y) {
    $angle = deg2rad(-$angle);
    $cos = cos($angle);
    $sin = sin($angle);
    $x = $x1 * $cos - $y1 * $sin;
    $y = $x1 * $sin + $y1 * $cos;
  }

}

/**
 * Class DrawBarcodeModelI25.
 */
class DrawBarcodeModelI25 {
  private static $encoding = array(
    'NNWWN',
    'WNNNW',
    'NWNNW',
    'WWNNN',
    'NNWNW',
    'WNWNN',
    'NWWNN',
    'NNNWW',
    'WNNWN',
    'NWNWN',
  );

  /**
   * Compute the barcode code value.
   *
   * @param string $code
   *   Code variable.
   * @param string $crc
   *   Crc variable.
   * @param string $type
   *   Type variable.
   *
   * @return string
   *   Return string variable.
   */
  public static function compute($code, $crc, $type) {
    if (!$crc) {
      if (strlen($code) % 2) {
        $code = '0' . $code;
      }
    }
    else {
      if (($type == 'int25') && (strlen($code) % 2 == 0)) {
        $code = '0' . $code;
      }
      $odd = TRUE;
      $sum = 0;
      for ($i = (strlen($code) - 1); $i > -1; $i--) {
        $v = intval($code[$i]);
        $sum += $odd ? 3 * $v : $v;
        $odd = !$odd;
      }
      $code .= (string) ((10 - $sum % 10) % 10);
    }
    return ($code);
  }

  /**
   * Get barcode digits.
   *
   * @param string $code
   *   Code variable.
   * @param string $crc
   *   Crc variable.
   * @param string $type
   *   Type variable.
   *
   * @return multitypestring
   *   Return string array.
   */
  public static function getDigit($code, $crc, $type) {
    $code = self::compute($code, $crc, $type);
    if ($code == '') {
      return ($code);
    }
    $result = '';
    if ($type == 'int25') {
      $result .= '1010';
      $end = strlen($code) / 2;
      for ($i = 0; $i < $end; $i++) {
        $c1 = $code[2 * $i];
        $c2 = $code[2 * $i + 1];
        for ($j = 0; $j < 5; $j++) {
          $result .= '1';
          if (self::$encoding[$c1][$j] == 'W') {
            $result .= '1';
          }
          $result .= '0';
          if (self::$encoding[$c2][$j] == 'W') {
            $result .= '0';
          }
        }
      }
      $result .= '1101';
    }
    elseif ($type == 'std25') {
      $result .= '11011010';
      $end = strlen($code);
      for ($i = 0; $i < $end; $i++) {
        $c = $code[$i];
        for ($j = 0; $j < 5; $j++) {
          $result .= '1';
          if (self::$encoding[$c][$j] == 'W') {
            $result .= '11';
          }
          $result .= '0';
        }
      }
      $result .= '11010110';
    }
    return ($result);
  }

}

/**
 * Class DrawBarcodeModelEAN.
 */
class DrawBarcodeModelEAN {
  private static $encoding = array(
    array('0001101', '0100111', '1110010'),
    array('0011001', '0110011', '1100110'),
    array('0010011', '0011011', '1101100'),
    array('0111101', '0100001', '1000010'),
    array('0100011', '0011101', '1011100'),
    array('0110001', '0111001', '1001110'),
    array('0101111', '0000101', '1010000'),
    array('0111011', '0010001', '1000100'),
    array('0110111', '0001001', '1001000'),
    array('0001011', '0010111', '1110100'),
  );

  private static $first = array('000000',
    '001011',
    '001101',
    '001110',
    '010011',
    '011001',
    '011100',
    '010101',
    '010110',
    '011010',
  );

  /**
   * Get code digit values.
   *
   * @param string $code
   *   Code variable.
   * @param string $type
   *   Type variable.
   *
   * @return string
   *   Return string variable.
   */
  public static function getDigit($code, $type) {
    $len = $type == 'ean8' ? 7 : 12;
    $code = substr($code, 0, $len);
    if (!preg_match('`[0-9]{' . $len . '}`', $code)) {
      return ('');
    }
    $code = self::compute($code, $type);
    $result = '101';
    if ($type == 'ean8') {
      for ($i = 0; $i < 4; $i++) {
        $result .= self::$encoding[intval($code[$i])][0];
      }
      $result .= '01010';
      for ($i = 4; $i < 8; $i++) {
        $result .= self::$encoding[intval($code[$i])][2];
      }
    }
    else {
      $seq = self::$first[intval($code[0])];
      for ($i = 1; $i < 7; $i++) {
        $result .= self::$encoding[intval($code[$i])][intval($seq[$i - 1])];
      }
      $result .= '01010';
      for ($i = 7; $i < 13; $i++) {
        $result .= self::$encoding[intval($code[$i])][2];
      }
    }
    $result .= '101';
    return ($result);
  }

  /**
   * Compute the barcode.
   *
   * @param string $code
   *   Code variable.
   * @param string $type
   *   Type variable.
   *
   * @return string
   *   Return string variable.
   */
  public static function compute($code, $type) {
    $len = $type == 'ean13' ? 12 : 7;
    $code = substr($code, 0, $len);
    if (!preg_match('`[0-9]{' . $len . '}`', $code)) {
      return ('');
    }
    $sum = 0;
    $odd = TRUE;
    for ($i = $len - 1; $i > -1; $i--) {
      $sum += ($odd ? 3 : 1) * intval($code[$i]);
      $odd = !$odd;
    }
    return ($code . ((string) ((10 - $sum % 10) % 10)));
  }

}

/**
 * Class DrawBarcodeModelMSI.
 */
class DrawBarcodeModelMSI {
  private static $encoding = array(
    '100100100100',
    '100100100110',
    '100100110100',
    '100100110110',
    '100110100100',
    '100110100110',
    '100110110100',
    '100110110110',
    '110100100100',
    '110100100110',
  );

  /**
   * Compute the barcode values.
   *
   * @param string $code
   *   Code variable.
   * @param string $crc
   *   Crc variable.
   *
   * @return string
   *   Return string variable.
   */
  public static function compute($code, $crc) {
    if (is_array($crc)) {
      if ($crc['crc1'] == 'mod10') {
        $code = self::computeMod10($code);
      }
      elseif ($crc['crc1'] == 'mod11') {
        $code = self::computeMod11($code);
      }
      if ($crc['crc2'] == 'mod10') {
        $code = self::computeMod10($code);
      }
      elseif ($crc['crc2'] == 'mod11') {
        $code = self::computeMod11($code);
      }
    }
    elseif ($crc) {
      $code = self::computeMod10($code);
    }
    return ($code);
  }

  /**
   * Compute the barcode values by Mod 10.
   *
   * @param string $code
   *   Code string variable.
   *
   * @return string
   *   Return string variable.
   */
  private static function computeMod10($code) {
    $len = strlen($code);
    $topart1 = $len % 2;
    $n1 = 0;
    $sum = 0;
    for ($i = 0; $i < $len; $i++) {
      if ($topart1) {
        $n1 = 10 * $n1 + intval($code[$i]);
      }
      else {
        $sum += intval($code[$i]);
      }
      $topart1 = !$topart1;
    }
    $s1 = (string) (2 * $n1);
    $len = strlen($s1);
    for ($i = 0; $i < $len; $i++) {
      $sum += intval($s1[$i]);
    }
    return ($code . ((string) (10 - $sum % 10) % 10));
  }

  /**
   * Compute the barcode values by Mod 11.
   *
   * @param string $code
   *   Code variable.
   *
   * @return string
   *   Return string variable.
   */
  private static function computeMod11($code) {
    $sum = 0;
    $weight = 2;
    for ($i = (strlen($code) - 1); $i > -1; $i--) {
      $sum += $weight * intval($code[$i]);
      $weight = $weight == 7 ? 2 : $weight + 1;
    }
    return ($code . ((string) (11 - $sum % 11) % 11));
  }

  /**
   * Get barcode digit values.
   *
   * @param string $code
   *   Code variable.
   * @param string $crc
   *   Crc variable.
   *
   * @return string
   *   Retrun string variable.
   */
  public static function getDigit($code, $crc) {
    if (preg_match('`[^0-9]`', $code)) {
      return '';
    }
    $result = '';
    $code = self::compute($code, FALSE);
    $result = '110';
    $len = strlen($code);
    for ($i = 0; $i < $len; $i++) {
      $result .= self::$encoding[intval($code[$i])];
    }
    $result .= '1001';
    return ($result);
  }

}

/**
 * Class DrawBarcodeModel11.
 */
class DrawBarcodeModel11 {
  private static $encoding = array(
    '101011',
    '1101011',
    '1001011',
    '1100101',
    '1011011',
    '1101101',
    '1001101',
    '1010011',
    '1101001',
    '110101',
    '101101',
  );

  /**
   * Get barcode dight values.
   *
   * @param string $code
   *   Code variable.
   *
   * @return string
   *   Return string variable.
   */
  public static function getDigit($code) {
    if (preg_match('`[^0-9\-]`', $code)) {
      return '';
    }
    $result = '';
    $intercharacter = '0';
    $result = '1011001' . $intercharacter;
    $len = strlen($code);
    for ($i = 0; $i < $len; $i++) {
      $index = $code[$i] == '-' ? 10 : intval($code[$i]);
      $result .= self::$encoding[$index] . $intercharacter;
    }
    $weightc    = 0;
    $weightsumc = 0;
    $weightk    = 1;
    $weightsumk = 0;
    for ($i = $len - 1; $i > -1; $i--) {
      $weightc = $weightc == 10 ? 1 : $weightc + 1;
      $weightk = $weightk == 10 ? 1 : $weightk + 1;
      $index = $code[$i] == '-' ? 10 : intval($code[$i]);
      $weightsumc += $weightc * $index;
      $weightsumk += $weightk * $index;
    }
    $c = $weightsumc % 11;
    $weightsumk += $c;
    $k = $weightsumk % 11;
    $result .= self::$encoding[$c] . $intercharacter;
    if ($len >= 10) {
      $result .= self::$encoding[$k] . $intercharacter;
    }
    $result = $result . '1011001';
    return ($result);
  }

}

/**
 * Class DrawBarcodeModel39.
 */
class DrawBarcodeModel39 {
  private static $encoding = array(
    '101001101101',
    '110100101011',
    '101100101011',
    '110110010101',
    '101001101011',
    '110100110101',
    '101100110101',
    '101001011011',
    '110100101101',
    '101100101101',
    '110101001011',
    '101101001011',
    '110110100101',
    '101011001011',
    '110101100101',
    '101101100101',
    '101010011011',
    '110101001101',
    '101101001101',
    '101011001101',
    '110101010011',
    '101101010011',
    '110110101001',
    '101011010011',
    '110101101001',
    '101101101001',
    '101010110011',
    '110101011001',
    '101101011001',
    '101011011001',
    '110010101011',
    '100110101011',
    '110011010101',
    '100101101011',
    '110010110101',
    '100110110101',
    '100101011011',
    '110010101101',
    '100110101101',
    '100100100101',
    '100100101001',
    '100101001001',
    '101001001001',
    '100101101101',
  );

  /**
   * Get barcode digit values.
   *
   * @param string $code
   *   Code variable.
   *
   * @return multitypestring
   *   Return string array variable.
   */
  public static function getDigit($code) {
    $table = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ-. $/+%*';
    $result = '';
    $intercharacter = '0';
    if (strpos($code, '*') !== FALSE) {
      return ('');
    }
    $code = strtoupper('*' . $code . '*');
    $len = strlen($code);
    for ($i = 0; $i < $len; $i++) {
      $index = strpos($table, $code[$i]);
      if ($index === FALSE) {
        return ('');
      }
      if ($i > 0) {
        $result .= $intercharacter;
      }
      $result .= self::$encoding[$index];
    }
    return ($result);
  }

}

/**
 * Class DrawBarcodeModel93.
 */
class DrawBarcodeModel93 {
  private static $encoding = array(
    '100010100',
    '101001000',
    '101000100',
    '101000010',
    '100101000',
    '100100100',
    '100100010',
    '101010000',
    '100010010',
    '100001010',
    '110101000',
    '110100100',
    '110100010',
    '110010100',
    '110010010',
    '110001010',
    '101101000',
    '101100100',
    '101100010',
    '100110100',
    '100011010',
    '101011000',
    '101001100',
    '101000110',
    '100101100',
    '100010110',
    '110110100',
    '110110010',
    '110101100',
    '110100110',
    '110010110',
    '110011010',
    '101101100',
    '101100110',
    '100110110',
    '100111010',
    '100101110',
    '111010100',
    '111010010',
    '111001010',
    '101101110',
    '101110110',
    '110101110',
    '100100110',
    '111011010',
    '111010110',
    '100110010',
    '101011110',
  );

  /**
   * Get barcode dight values.
   *
   * @param string $code
   *   Code variable.
   * @param string $crc
   *   Crc variable.
   *
   * @return string
   *   Return string variable.
   */
  public static function getDigit($code, $crc) {
    $table = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ-. $/+%____*';
    $result = '';
    if (strpos($code, '*') !== FALSE) {
      return ('');
    }
    $code = strtoupper($code);
    $result .= self::$encoding[47];
    $len = strlen($code);
    for ($i = 0; $i < $len; $i++) {
      $c = $code[$i];
      $index = strpos($table, $c);
      if (($c == '_') || ($index === FALSE)) {
        return ('');
      }
      $result .= self::$encoding[$index];
    }
    if ($crc) {
      $weightc = 0;
      $weightsumc = 0;
      $weightk = 1;
      $weightsumk = 0;
      for ($i = $len - 1; $i > -1; $i--) {
        $weightc = $weightc == 20 ? 1 : $weightc + 1;
        $weightk = $weightk == 15 ? 1 : $weightk + 1;
        $index = strpos($table, $code[$i]);
        $weightsumc += $weightc * $index;
        $weightsumk += $weightk * $index;
      }
      $c = $weightsumc % 47;
      $weightsumk += $c;
      $k = $weightsumk % 47;
      $result .= self::$encoding[$c];
      $result .= self::$encoding[$k];
    }
    $result .= self::$encoding[47];
    $result .= '1';
    return ($result);
  }

}

/**
 * Class DrawBarcodeModel128.
 */
class DrawBarcodeModel128 {
  private static $encoding = array(
    '11011001100',
    '11001101100',
    '11001100110',
    '10010011000',
    '10010001100',
    '10001001100',
    '10011001000',
    '10011000100',
    '10001100100',
    '11001001000',
    '11001000100',
    '11000100100',
    '10110011100',
    '10011011100',
    '10011001110',
    '10111001100',
    '10011101100',
    '10011100110',
    '11001110010',
    '11001011100',
    '11001001110',
    '11011100100',
    '11001110100',
    '11101101110',
    '11101001100',
    '11100101100',
    '11100100110',
    '11101100100',
    '11100110100',
    '11100110010',
    '11011011000',
    '11011000110',
    '11000110110',
    '10100011000',
    '10001011000',
    '10001000110',
    '10110001000',
    '10001101000',
    '10001100010',
    '11010001000',
    '11000101000',
    '11000100010',
    '10110111000',
    '10110001110',
    '10001101110',
    '10111011000',
    '10111000110',
    '10001110110',
    '11101110110',
    '11010001110',
    '11000101110',
    '11011101000',
    '11011100010',
    '11011101110',
    '11101011000',
    '11101000110',
    '11100010110',
    '11101101000',
    '11101100010',
    '11100011010',
    '11101111010',
    '11001000010',
    '11110001010',
    '10100110000',
    '10100001100',
    '10010110000',
    '10010000110',
    '10000101100',
    '10000100110',
    '10110010000',
    '10110000100',
    '10011010000',
    '10011000010',
    '10000110100',
    '10000110010',
    '11000010010',
    '11001010000',
    '11110111010',
    '11000010100',
    '10001111010',
    '10100111100',
    '10010111100',
    '10010011110',
    '10111100100',
    '10011110100',
    '10011110010',
    '11110100100',
    '11110010100',
    '11110010010',
    '11011011110',
    '11011110110',
    '11110110110',
    '10101111000',
    '10100011110',
    '10001011110',
    '10111101000',
    '10111100010',
    '11110101000',
    '11110100010',
    '10111011110',
    '10111101110',
    '11101011110',
    '11110101110',
    '11010000100',
    '11010010000',
    '11010011100',
    '11000111010',
  );

  /**
   * Get barcode dight values.
   *
   * @param string $code
   *   Code variable.
   *
   * @return string
   *   Return string variable.
   */
  public static function getDigit($code) {
    $tableb = " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~";
    $result = "";
    $sum = 0;
    $isum = 0;
    $i = 0;
    $j = 0;
    $value = 0;
    $len = strlen($code);
    for ($i = 0; $i < $len; $i++) {
      if (strpos($tableb, $code[$i]) === FALSE) {
        return ("");
      }
    }
    $tablecactivated = $len > 1;
    for ($i = 0; $i < 3 && $i < $len; $i++) {
      $tablecactivated &= preg_match('`[0-9]`', $code[$i]);
    }
    $sum = $tablecactivated ? 105 : 104;
    $result = self::$encoding[$sum];
    $i = 0;
    while ($i < $len) {
      if (!$tablecactivated) {
        $j = 0;
        while (($i + $j < $len) && preg_match('`[0-9]`', $code[$i + $j])) {
          $j++;
        }
        $tablecactivated = ($j > 5) || (($i + $j - 1 == $len) && ($j > 3));
        if ($tablecactivated) {
          $result .= self::$encoding[99];
          $sum += ++$isum * 99;
        }
      }
      elseif (($i == $len - 1) || (preg_match('`[^0-9]`', $code[$i])) || (preg_match('`[^0-9]`', $code[$i + 1]))) {
        $tablecactivated = FALSE;
        $result .= self::$encoding[100];
        $sum += ++$isum * 100;
      }
      if ($tablecactivated) {
        $value = intval(substr($code, $i, 2));
        $i += 2;
      }
      else {
        $value = strpos($tableb, $code[$i]);
        $i++;
      }
      $result .= self::$encoding[$value];
      $sum += ++$isum * $value;
    }
    $result .= self::$encoding[$sum % 103];
    $result .= self::$encoding[106];
    $result .= '11';
    return ($result);
  }

}

/**
 * Class DrawBarcodeModelCodabar.
 */
class DrawBarcodeModelCodabar {
  private static $encoding = array(
    '101010011',
    '101011001',
    '101001011',
    '110010101',
    '101101001',
    '110101001',
    '100101011',
    '100101101',
    '100110101',
    '110100101',
    '101001101',
    '101100101',
    '1101011011',
    '1101101011',
    '1101101101',
    '1011011011',
    '1011001001',
    '1010010011',
    '1001001011',
    '1010011001',
  );

  /**
   * Get barcode dight values.
   *
   * @param string $code
   *   Code variable.
   *
   * @return multitypestring
   *   Return string array variable.
   */
  public static function getDigit($code) {
    $table = '0123456789-$:/.+';
    $result = '';
    $intercharacter = '0';
    $result .= self::$encoding[16] . $intercharacter;
    $len = strlen($code);
    for ($i = 0; $i < $len; $i++) {
      $index = strpos($table, $code[$i]);
      if ($index === FALSE) {
        return ('');
      }
      $result .= self::$encoding[$index] . $intercharacter;
    }
    $result .= self::$encoding[16];
    return ($result);
  }

}

/**
 * Class DrawBarcodeModelDatamatrix.
 */
class DrawBarcodeModelDatamatrix {
  private static $encoding = array(
    '101010011',
    '101011001',
    '101001011',
    '110010101',
    '101101001',
    '110101001',
    '100101011',
    '100101101',
    '100110101',
    '110100101',
    '101001101',
    '101100101',
    '1101011011',
    '1101101011',
    '1101101101',
    '1011011011',
    '1011001001',
    '1010010011',
    '1001001011',
    '1010011001',
  );
  private static $lengthRows = array(
    10, 12, 14, 16, 18, 20, 22, 24, 26,
    32, 36, 40, 44, 48, 52, 64, 72, 80, 88, 96, 104, 120, 132, 144,
    8, 8, 12, 12, 16, 16,
  );
  private static $lengthCols = array(
    10, 12, 14, 16, 18, 20, 22, 24, 26,
    32, 36, 40, 44, 48, 52, 64, 72, 80, 88, 96, 104, 120, 132, 144,
    18, 32, 26, 36, 36, 48,
  );
  private static $mappingRows = array(
    8, 10, 12, 14, 16, 18, 20, 22, 24,
    28, 32, 36, 40, 44, 48, 56, 64, 72, 80, 88, 96, 108, 120, 132,
    6, 6, 10, 10, 14, 14,
  );
  private static $mappingCols = array(
    8, 10, 12, 14, 16, 18, 20, 22, 24,
    28, 32, 36, 40, 44, 48, 56, 64, 72,
    80, 88, 96, 108, 120, 132,
    16, 28, 24, 32, 32, 44,
  );
  private static $dataCwCount = array(
    3, 5, 8, 12, 18, 22, 30, 36,
    44, 62, 86, 114, 144, 174, 204, 280,
    368, 456, 576, 696, 816, 1050,
    1304, 1558, 5, 10, 16, 22, 32, 49,
  );
  private static $solomonCwCount = array(
    5, 7, 10, 12, 14, 18, 20, 24, 28,
    36, 42, 48, 56, 68, 84, 112, 144, 192,
    224, 272, 336, 408, 496, 620,
    7, 11, 14, 18, 24, 28,
  );
  private static $dataRegionRows = array(
    8, 10, 12, 14, 16, 18, 20, 22,
    24, 14, 16, 18, 20, 22, 24, 14, 16,
    18, 20, 22, 24, 18, 20, 22,
    6, 6, 10, 10, 14, 14,
  );
  private static $dataRegionCols = array(
    8, 10, 12, 14, 16, 18, 20, 22,
    24, 14, 16, 18, 20, 22, 24, 14, 16,
    18, 20, 22, 24, 18, 20, 22,
    16, 14, 24, 16, 16, 22,
  );
  private static $regionRows = array(
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 2, 2, 2, 2, 2, 2, 4, 4, 4, 4, 4, 4, 6, 6, 6,
    1, 1, 1, 1, 1, 1,
  );
  private static $regionCols = array(
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 2, 2, 2, 2, 2, 2, 4, 4, 4, 4, 4, 4, 6, 6, 6,
    1, 2, 1, 2, 2, 2,
  );
  private static $interleavedBlocks = array(
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 2, 2, 4, 4, 4, 4, 6, 6, 8, 8,
    1, 1, 1, 1, 1, 1,
  );
  private static $logTab = array(
    -255, 255, 1, 240, 2, 225, 241, 53, 3,
    38, 226, 133, 242, 43, 54, 210, 4, 195, 39, 114,
    227, 106, 134, 28, 243, 140, 44, 23, 55, 118, 211, 234,
    5, 219, 196, 96, 40, 222, 115, 103, 228, 78, 107, 125,
    135, 8, 29, 162, 244, 186, 141, 180, 45, 99, 24, 49, 56,
    13, 119, 153, 212, 199, 235, 91, 6, 76, 220, 217, 197,
    11, 97, 184, 41, 36, 223, 253, 116, 138, 104, 193, 229,
    86, 79, 171, 108, 165, 126, 145, 136, 34, 9, 74, 30,
    32, 163, 84, 245, 173, 187, 204, 142, 81, 181, 190, 46,
    88, 100, 159, 25, 231, 50, 207, 57, 147, 14, 67, 120,
    128, 154, 248, 213, 167, 200, 63, 236, 110, 92, 176, 7,
    161, 77, 124, 221, 102, 218, 95, 198, 90, 12, 152,
    98, 48, 185, 179, 42, 209, 37, 132, 224, 52, 254,
    239, 117, 233, 139, 22, 105, 27, 194, 113, 230, 206,
    87, 158, 80, 189, 172, 203, 109, 175, 166, 62, 127, 247,
    146, 66, 137, 192, 35, 252, 10, 183, 75, 216, 31, 83, 33, 73,
    164, 144, 85, 170, 246, 65, 174, 61, 188, 202, 205,
    157, 143, 169, 82, 72, 182, 215, 191, 251, 47, 178, 89,
    151, 101, 94, 160, 123, 26, 112, 232, 21, 51, 238, 208,
    131, 58, 69, 148, 18, 15, 16, 68, 17, 121, 149,
    129, 19, 155, 59, 249, 70, 214, 250, 168, 71, 201,
    156, 64, 60, 237, 130, 111, 20, 93, 122, 177, 150,
  );
  private static $aLogTab = array(
    1, 2, 4, 8, 16, 32, 64, 128, 45, 90,
    180, 69, 138, 57, 114, 228, 229, 231, 227, 235, 251, 219, 155, 27, 54,
    108, 216, 157, 23, 46, 92, 184, 93, 186, 89, 178, 73, 146, 9, 18, 36,
    72, 144, 13, 26, 52, 104, 208, 141, 55, 110, 220, 149, 7, 14, 28, 56,
    112, 224, 237, 247, 195, 171, 123, 246, 193, 175, 115, 230, 225, 239,
    243, 203, 187, 91, 182, 65, 130, 41, 82, 164, 101, 202, 185, 95, 190,
    81, 162, 105, 210, 137, 63, 126, 252, 213, 135, 35, 70, 140, 53, 106,
    212, 133, 39, 78, 156, 21, 42, 84, 168, 125, 250, 217, 159, 19, 38, 76,
    152, 29, 58, 116, 232, 253, 215, 131, 43, 86, 172, 117, 234, 249, 223,
    147, 11, 22, 44, 88, 176, 77, 154, 25, 50, 100, 200, 189, 87, 174, 113,
    226, 233, 255, 211, 139, 59, 118, 236, 245, 199, 163, 107, 214, 129,
    47, 94, 188, 85, 170, 121, 242, 201, 191, 83, 166, 97, 194, 169, 127,
    254, 209, 143, 51, 102, 204, 181, 71, 142, 49, 98, 196, 165, 103, 206,
    177, 79, 158, 17, 34, 68, 136, 61, 122, 244, 197, 167, 99, 198, 161,
    111, 222, 145, 15, 30, 60, 120, 240, 205, 183, 67, 134, 33, 66, 132,
    37, 74, 148, 5, 10, 20, 40, 80, 160, 109, 218, 153, 31, 62, 124, 248,
    221, 151, 3, 6, 12, 24, 48, 96, 192, 173, 119, 238, 241, 207, 179, 75,
    150, 1,
  );

  /**
   * ChampGaloisMult for Log formula.
   *
   * @param int $a
   *   A integer variable.
   * @param int $b
   *   B integer variable.
   *
   * @return multitypeint
   *   Return integer variable.
   */
  private static function champGaloisMult($a, $b) {
    if (!$a || !$b) {
      return 0;
    }
    return self::$aLogTab[(self::$logTab[$a] + self::$logTab[$b]) % 255];
  }

  /**
   * ChampGaloisDoub for log formula.
   *
   * @param int $a
   *   A integer variable.
   * @param int $b
   *   B integer variable.
   *
   * @return multitypeint
   *   Return integer variable.
   */
  private static function champGaloisDoub($a, $b) {
    if (!$a) {
      return 0;
    }
    if (!$b) {
      return $a;
    }
    return self::$aLogTab[(self::$logTab[$a] + $b) % 255];
  }

  /**
   * ChampGaloisSum for log formula.
   *
   * @param int $a
   *   A integer variable.
   * @param int $b
   *   B integer variable.
   *
   * @return multitypeint
   *   Return bool variable.
   */
  private static function champGaloisSum($a, $b) {
    return $a ^ $b;
  }

  /**
   * Select the barcode index values.
   *
   * @param int $data_code_words_count
   *   Data code words count variable.
   * @param int $rectangular
   *   Rectangular variable.
   *
   * @return int
   *   Return integer variable.
   */
  private static function selectIndex($data_code_words_count, $rectangular) {
    if (($data_code_words_count < 1 || $data_code_words_count > 1558) && !$rectangular) {
      return -1;
    }
    if (($data_code_words_count < 1 || $data_code_words_count > 49) && $rectangular) {
      return -1;
    }
    $n = $rectangular ? 24 : 0;
    while (self::$data_cw_count[$n] < $data_code_words_count) {
      $n++;
    }
    return $n;
  }

  /**
   * Encode the input data code to ASCII values.
   *
   * @param string $text
   *   Text variable.
   *
   * @return multitypeint
   *   Return integer variable.
   */
  private static function encodeDataCodeWordsAscii($text) {
    $data_code_words = array();
    $n = 0;
    $len = strlen($text);
    for ($i = 0; $i < $len; $i++) {
      $c = ord($text[$i]);
      if ($c > 127) {
        $data_code_words[$n] = 235;
        $c -= 127;
        $n++;
      }
      elseif (($c >= 48 && $c <= 57) && ($i + 1 < $len) && (preg_match('`[0-9]`', $text[$i + 1]))) {
        $c = (($c - 48) * 10) + intval($text[$i + 1]);
        $c += 130;
        $i++;
      }
      else {
        $c++;
      }
      $data_code_words[$n] = $c;
      $n++;
    }
    return $data_code_words;
  }

  /**
   * Add PAD CW array.
   *
   * @param string $tab
   *   Tab variable.
   * @param string $from
   *   From variable.
   * @param string $to
   *   To variable.
   */
  private static function addPadCw(&$tab, $from, $to) {
    if ($from >= $to) {
      return;
    }
    $tab[$from] = 129;
    for ($i = $from + 1; $i < $to; $i++) {
      $r = ((149 * ($i + 1)) % 253) + 1;
      $tab[$i] = (129 + $r) % 254;
    }
  }

  /**
   * Calculate SolFator table values.
   *
   * @param string $solomon_cw_count
   *   Solomon Cw count variable.
   *
   * @return string
   *   Return array variable.
   */
  private static function calculSolFactorTable($solomon_cw_count) {
    $g = array_fill(0, $solomon_cw_count + 1, 1);
    for ($i = 1; $i <= $solomon_cw_count; $i++) {
      for ($j = $i - 1; $j >= 0; $j--) {
        $g[$j] = self::champGaloisDoub($g[$j], $i);
        if ($j > 0) {
          $g[$j] = self::champGaloisSum($g[$j], $g[$j - 1]);
        }
      }
    }
    return $g;
  }

  /**
   * Add Reed SolomonCW values.
   *
   * @param string $n_solomon_cw
   *   N Solomon cw variable.
   * @param string $coeff_tab
   *   Coeff tab variable.
   * @param string $n_data_cw
   *   N data cw variable.
   * @param string $data_tab
   *   Data tabe variable.
   * @param string $blocks
   *   Blocks variable.
   *
   * @return string
   *   Return array variable.
   */
  private static function addReedSolomonCw($n_solomon_cw, $coeff_tab, $n_data_cw, &$data_tab, $blocks) {
    $temp = 0;
    $error_blocks = $n_solomon_cw / $blocks;
    $correction_cw = array();
    for ($k = 0; $k < $blocks; $k++) {
      for ($i = 0; $i < $error_blocks; $i++) {
        $correction_cw[$i] = 0;
      }
      for ($i = $k; $i < $n_data_cw; $i += $blocks) {
        $temp = self::champGaloisSum($data_tab[$i], $correction_cw[$error_blocks - 1]);
        for ($j = $error_blocks - 1; $j >= 0; $j--) {
          if (!$temp) {
            $correction_cw[$j] = 0;
          }
          else {
            $correction_cw[$j] = self::champGaloisMult($temp, $coeff_tab[$j]);
          }
          if ($j > 0) {
            $correction_cw[$j] = self::champGaloisSum($correction_cw[$j - 1], $correction_cw[$j]);
          }
        }
      }
      $j = $n_data_cw + $k;
      for ($i = $error_blocks - 1; $i >= 0; $i--) {
        $data_tab[$j] = $correction_cw[$i];
        $j = $j + $blocks;
      }
    }
    return $data_tab;
  }

  /**
   * Get Bit values.
   *
   * @param int $entier
   *   Entier variable.
   *
   * @return multitypeint
   *   Return int array variable.
   */
  private static function getBits($entier) {
    $bits = array();
    for ($i = 0; $i < 8; $i++) {
      $bits[$i] = $entier & (128 >> $i) ? 1 : 0;
    }
    return $bits;
  }

  /**
   * Next function to move to barcode next index.
   *
   * @param int $etape
   *   Etape variable.
   * @param int $total_rows
   *   Total rows variable.
   * @param int $total_cols
   *   Total cols variable.
   * @param int $code_words_bits
   *   Code words bits variable.
   * @param int $datamatrix
   *   Data matrix variable.
   * @param int $assigned
   *   Assigned variable.
   */
  private static function next($etape, $total_rows, $total_cols, $code_words_bits, &$datamatrix, &$assigned) {
    $chr = 0;
    $row = 4;
    $col = 0;
    do {
      if (($row == $total_rows) && ($col == 0)) {
        self::patternShapeSpecial1($datamatrix, $assigned, $code_words_bits[$chr], $total_rows, $total_cols);
        $chr++;
      }
      elseif (($etape < 3) && ($row == $total_rows - 2) && ($col == 0) && ($total_cols % 4 != 0)) {
        self::patternShapeSpecial2($datamatrix, $assigned, $code_words_bits[$chr], $total_rows, $total_cols);
        $chr++;
      }
      elseif (($row == $total_rows - 2) && ($col == 0) && ($total_cols % 8 == 4)) {
        self::patternShapeSpecial3($datamatrix, $assigned, $code_words_bits[$chr], $total_rows, $total_cols);
        $chr++;
      }
      elseif (($row == $total_rows + 4) && ($col == 2) && ($total_cols % 8 == 0)) {
        self::patternShapeSpecial4($datamatrix, $assigned, $code_words_bits[$chr], $total_rows, $total_cols);
        $chr++;
      }
      do {
        if (($row < $total_rows) && ($col >= 0) && (!isset($assigned[$row][$col]) || $assigned[$row][$col] != 1)) {
          self::patternShapeStandard($datamatrix, $assigned, $code_words_bits[$chr], $row, $col, $total_rows, $total_cols);
          $chr++;
        }
        $row -= 2;
        $col += 2;
      } while (($row >= 0) && ($col < $total_cols));
      $row += 1;
      $col += 3;
      do {
        if (($row >= 0) && ($col < $total_cols) && (!isset($assigned[$row][$col]) || $assigned[$row][$col] != 1)) {
          self::patternShapeStandard($datamatrix, $assigned, $code_words_bits[$chr], $row, $col, $total_rows, $total_cols);
          $chr++;
        }
        $row += 2;
        $col -= 2;
      } while (($row < $total_rows) && ($col >= 0));
      $row += 3;
      $col += 1;
    } while (($row < $total_rows) || ($col < $total_cols));
  }

  /**
   * Shape pattern in a standard mode.
   *
   * @param int $datamatrix
   *   Data matrix variable.
   * @param int $assigned
   *   Assigned variable.
   * @param int $bits
   *   Bits variable.
   * @param int $row
   *   Row variable.
   * @param int $col
   *   Column variable.
   * @param int $total_rows
   *   Total rows variable.
   * @param int $total_cols
   *   Total cols variable.
   */
  private static function patternShapeStandard(&$datamatrix, &$assigned, $bits, $row, $col, $total_rows, $total_cols) {
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[0], $row - 2, $col - 2, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[1], $row - 2, $col - 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[2], $row - 1, $col - 2, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[3], $row - 1, $col - 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[4], $row - 1, $col, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[5], $row, $col - 2, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[6], $row, $col - 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[7], $row, $col, $total_rows, $total_cols);
  }

  /**
   * Shape the pattern in Special1 mode.
   *
   * @param int $datamatrix
   *   Data matrix variable.
   * @param int $assigned
   *   Assigned variable.
   * @param int $bits
   *   Bits variable.
   * @param int $total_rows
   *   Total rows variable.
   * @param int $total_cols
   *   Total cols variable.
   */
  private static function patternShapeSpecial1(&$datamatrix, &$assigned, $bits, $total_rows, $total_cols) {
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[0], $total_rows - 1, 0, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[1], $total_rows - 1, 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[2], $total_rows - 1, 2, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[3], 0, $total_cols - 2, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[4], 0, $total_cols - 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[5], 1, $total_cols - 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[6], 2, $total_cols - 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[7], 3, $total_cols - 1, $total_rows, $total_cols);
  }

  /**
   * Shape the pattern in Special2 mode.
   *
   * @param int $datamatrix
   *   Data matrix variable.
   * @param int $assigned
   *   Assigned variable.
   * @param int $bits
   *   Bits variable.
   * @param int $total_rows
   *   Total rows variable.
   * @param int $total_cols
   *   Total cols variable.
   */
  private static function patternShapeSpecial2(&$datamatrix, &$assigned, $bits, $total_rows, $total_cols) {
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[0], $total_rows - 3, 0, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[1], $total_rows - 2, 0, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[2], $total_rows - 1, 0, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[3], 0, $total_cols - 4, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[4], 0, $total_cols - 3, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[5], 0, $total_cols - 2, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[6], 0, $total_cols - 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[7], 1, $total_cols - 1, $total_rows, $total_cols);
  }

  /**
   * Shape the pattern in Special3 mode.
   *
   * @param int $datamatrix
   *   Data matrix variable.
   * @param int $assigned
   *   Assigned variable.
   * @param int $bits
   *   Bits variable.
   * @param int $total_rows
   *   Total rows variable.
   * @param int $total_cols
   *   Total column variable.
   */
  private static function patternShapeSpecial3(&$datamatrix, &$assigned, $bits, $total_rows, $total_cols) {
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[0], $total_rows - 3, 0, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[1], $total_rows - 2, 0, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[2], $total_rows - 1, 0, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[3], 0, $total_cols - 2, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[4], 0, $total_cols - 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[5], 1, $total_cols - 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[6], 2, $total_cols - 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[7], 3, $total_cols - 1, $total_rows, $total_cols);
  }

  /**
   * Shape the pattern in Special4 mode.
   *
   * @param int $datamatrix
   *   Data matrix variable.
   * @param int $assigned
   *   Assigned variable.
   * @param int $bits
   *   Bits variable.
   * @param int $total_rows
   *   Total rows variable.
   * @param int $total_cols
   *   Total column variable.
   */
  private static function patternShapeSpecial4(&$datamatrix, &$assigned, $bits, $total_rows, $total_cols) {
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[0], $total_rows - 1, 0, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[1], $total_rows - 1, $total_cols - 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[2], 0, $total_cols - 3, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[3], 0, $total_cols - 2, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[4], 0, $total_cols - 1, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[5], 1, $total_cols - 3, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[6], 1, $total_cols - 2, $total_rows, $total_cols);
    self::placeBitInDatamatrix($datamatrix, $assigned, $bits[7], 1, $total_cols - 1, $total_rows, $total_cols);
  }

  /**
   * Place the Bit value in a Data matrix.
   *
   * @param int $datamatrix
   *   Data matrix variable.
   * @param int $assigned
   *   Assigned variable.
   * @param int $bit
   *   Bit variable.
   * @param int $row
   *   Row variable.
   * @param int $col
   *   Column variable.
   * @param int $total_rows
   *   Total rows variable.
   * @param int $total_cols
   *   Total column variable.
   */
  private static function placeBitInDatamatrix(&$datamatrix, &$assigned, $bit, $row, $col, $total_rows, $total_cols) {
    if ($row < 0) {
      $row += $total_rows;
      $col += 4 - (($total_rows + 4) % 8);
    }
    if ($col < 0) {
      $col += $total_cols;
      $row += 4 - (($total_cols + 4) % 8);
    }
    if (!isset($assigned[$row][$col]) || $assigned[$row][$col] != 1) {
      $datamatrix[$row][$col] = $bit;
      $assigned[$row][$col] = 1;
    }
  }

  /**
   * Add finder pattern to the barcode.
   *
   * @param int $datamatrix
   *   Data matrix variable.
   * @param int $rows_region
   *   Rows region variable.
   * @param int $cols_region
   *   Column region variable.
   * @param int $rows_region_cw
   *   Rows region cw variable.
   * @param int $cols_region_cw
   *   Column region cw variable.
   *
   * @return string
   *   Return array variable.
   */
  private static function addFinderPattern($datamatrix, $rows_region, $cols_region, $rows_region_cw, $cols_region_cw) {
    $total_rows_cw = ($rows_region_cw + 2) * $rows_region;
    $total_cols_cw = ($cols_region_cw + 2) * $cols_region;
    $datamatrix_temp = array();
    $datamatrix_temp[0] = array_fill(0, $total_cols_cw + 2, 0);
    for ($i = 0; $i < $total_rows_cw; $i++) {
      $datamatrix_temp[$i + 1] = array();
      $datamatrix_temp[$i + 1][0] = 0;
      $datamatrix_temp[$i + 1][$total_cols_cw + 1] = 0;
      for ($j = 0; $j < $total_cols_cw; $j++) {
        if ($i % ($rows_region_cw + 2) == 0) {
          if ($j % 2 == 0) {
            $datamatrix_temp[$i + 1][$j + 1] = 1;
          }
          else {
            $datamatrix_temp[$i + 1][$j + 1] = 0;
          }
        }
        elseif ($i % ($rows_region_cw + 2) == $rows_region_cw + 1) {
          $datamatrix_temp[$i + 1][$j + 1] = 1;
        }
        elseif ($j % ($cols_region_cw + 2) == $cols_region_cw + 1) {
          if ($i % 2 == 0) {
            $datamatrix_temp[$i + 1][$j + 1] = 0;
          }
          else {
            $datamatrix_temp[$i + 1][$j + 1] = 1;
          }
        }
        elseif ($j % ($cols_region_cw + 2) == 0) {
          $datamatrix_temp[$i + 1][$j + 1] = 1;
        }
        else {
          $datamatrix_temp[$i + 1][$j + 1] = 0;
          $datamatrix_temp[$i + 1][$j + 1] = $datamatrix[$i - 1 - (2 * (floor($i / ($rows_region_cw + 2))))][$j - 1 - (2 * (floor($j / ($cols_region_cw + 2))))];
        }
      }
    }
    $datamatrix_temp[$total_rows_cw + 1] = array();
    for ($j = 0; $j < $total_cols_cw + 2; $j++) {
      $datamatrix_temp[$total_rows_cw + 1][$j] = 0;
    }
    return $datamatrix_temp;
  }

  /**
   * Get barcode dight values.
   *
   * @param string $text
   *   Text  variable.
   * @param string $rectangular
   *   Rectangular variable.
   *
   * @return int
   *   Return digits property.
   */
  public static function getDigit($text, $rectangular) {
    $data_code_words = self::encodeDataCodeWordsAscii($text);
    $data_cw_count = count($data_code_words);
    $index = self::selectIndex($data_cw_count, $rectangular);
    $total_data_cw_count = self::$data_cw_count[$index];
    $solomon_cw_count = self::$solomon_cw_count[$index];
    $total_cw_count = $total_data_cw_count + $solomon_cw_count;
    $rows_total = self::$length_rows[$index];
    $cols_total = self::$length_cols[$index];
    $rows_region = self::$region_rows[$index];
    $cols_region = self::$region_cols[$index];
    $rows_region_cw = self::$data_region_rows[$index];
    $cols_region_cw = self::$data_region_cols[$index];
    $rows_length_matrice = $rows_total - 2 * $rows_region;
    $cols_length_matrice = $cols_total - 2 * $cols_region;
    $blocks = self::$interleaved_blocks[$index];
    $error_blocks = $solomon_cw_count / $blocks;
    self::addPadCw($data_code_words, $data_cw_count, $total_data_cw_count);
    $g = self::calculSolFactorTable($error_blocks);
    self::addReedSolomonCw($solomon_cw_count, $g, $total_data_cw_count, $data_code_words, $blocks);
    $code_words_bits = array();
    for ($i = 0; $i < $total_cw_count; $i++) {
      $code_words_bits[$i] = self::getBits($data_code_words[$i]);
    }
    $datamatrix = array_fill(0, $cols_length_matrice, array());
    $assigned = array_fill(0, $cols_length_matrice, array());
    if ((($rows_length_matrice * $cols_length_matrice) % 8) == 4) {
      $datamatrix[$rows_length_matrice - 2][$cols_length_matrice - 2] = 1;
      $datamatrix[$rows_length_matrice - 1][$cols_length_matrice - 1] = 1;
      $datamatrix[$rows_length_matrice - 1][$cols_length_matrice - 2] = 0;
      $datamatrix[$rows_length_matrice - 2][$cols_length_matrice - 1] = 0;
      $assigned[$rows_length_matrice - 2][$cols_length_matrice - 2] = 1;
      $assigned[$rows_length_matrice - 1][$cols_length_matrice - 1] = 1;
      $assigned[$rows_length_matrice - 1][$cols_length_matrice - 2] = 1;
      $assigned[$rows_length_matrice - 2][$cols_length_matrice - 1] = 1;
    }
    self::next(0, $rows_length_matrice, $cols_length_matrice, $code_words_bits, $datamatrix, $assigned);
    $datamatrix = self::addFinderPattern($datamatrix, $rows_region, $cols_region, $rows_region_cw, $cols_region_cw);
    return $datamatrix;
  }

}
